/*
 * @(#)MTFilter.java	1.00 99/03/15
 * 
 * Copyright (c) 1999 International Business Machines. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of IBM.
 * You shall not disclose such Confidential Information and shall use it
 * only in accordance with the terms of the license agreement you entered
 * into with IBM.
 * 
 * IBM MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. IBM SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * CopyrightVersion 1.0
 *
 *
 */

import java.io.*;
import java.util.Enumeration;
import java.util.StringTokenizer;

import javax.servlet.http.*;
import javax.servlet.*;

/**
 * This is an example of how one might write a Filter servlet that translates
 * the HTML text (generated from a servlet response) to a given target language.
 * The target language is determined by the ACCEPT_LANGUAGE HTTP Header Tag.
 * 
 * This servlet uses the IBM Machine Translation Java API's to translate to
 * the target language.
 *
 * Note: This servlet must be configured to "point" to the correct machine
 *       translation directory either using the "mtPath" servlet init parameter or
 *       by setting the static default "defaultLtPath".
 *
 * Note: This servlet should be made aware of UTF strings.
 *
 */
public class MTFilter extends HttpServlet
{

    // Strings
    final  static  String format        = "*format=html";   // MT option indicating HTML content translation
    final  static  String defaultMtPath = "C:\\Program Files\\IBM\\IBM Machine Translation\\";
    static String useLang;     // language to translate into, specified in admin
    static String mtPath;      // path where MT libraries can be found     

    // Target Language Tag, ClassName array
    // Note: As of WAS 3.0 the MT Engine supports German and Chinese Languages... However, this 
    //       Servlet is ready for the others when they become available
    // Note: To add a language, simply add entry to array.  The code in this class will 
    // dynamically handle the additional language.
    final static String[] ltLang = 
    {
        "cn",  //Simplified Chinese Engine");
        "da",  //Danish Engine
        "de",  //German Engine
        "es",  //Spanish Engine
        "fr",  //French Engine
        "it",  //Italian Engine
        "jp",  //Japanese Engine
        "kr",  //Korean Engine
        "nl",  //Dutch Engine
        "pl",  //Portuguese Engine
        "ru",  //Russian Engine
        "tw"   //Traditional Chinese Engine
    };

    // instance of a language translator class (corresponding to above array)
    static com.ibm.lt.LTinterface ltClass[] = new com.ibm.lt.LTinterface[ltLang.length];

    /**
     * init
     *
     * Initializes the servlet. Reads and processes parameters to servlet
     * This servlet supports two input parameters.  They are:
     *    mtPath   - The path to the MT libraries (E.g., \IBMHTTP\MT)
     *    useLang  - Forces a give language to be used (E.g., "de", which is German).
     *             - the value must match a value in the ltLang array defined above
     *
     */
    public void init(ServletConfig conf) throws ServletException 
    {
        super.init(conf);

        // This servlet supports an "mtPath" parameter which must be set up (via admin)
        // to override the default defaultMtPath
        if ( (mtPath = getInitParameter ("mtPath")) == null )
        {
            System.out.println("Machine Translate Filter (MTFilter): WARNING 'mtPath' param not set for this MTFilter instance. Using default value "+defaultMtPath);
            mtPath = defaultMtPath; // use default path
        }
        //  useLang  - Forces a give language to be used (E.g., "de", which is German).
        //  the value must match a value in the ltLang array defined above
        if ( (useLang = getInitParameter ("useLang")) == null )
        {
            System.out.println("Machine Translate Filter (MTFilter): WARNING 'useLang' param not set for this MTFilter instance. Using HTTP ACCEPT-LANGUAGE header tag for target language.");
            useLang = null;
        }
    }

    /**
     * service
     *
     * The Service Method is the primary entry point of Servlets.
     *
     * @param      req    the request object 
     * @param      res    the response object
     * @return     <code>void</void>
     * @exception  IOException problem reading from or writing to
     *             the req or res streams
     */
    public void service(HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException  
    {
        try
        {
            String contentType=null;                        // content type of request buffer
            int    langIndex=-1;                            // language index (into ltLang array)
            ServletOutputStream out = res.getOutputStream();// set up output stream
            ServletInputStream  in  = req.getInputStream(); // set up output stream

            // Copy headers from request buffer to response buffer...
            // While coping headers check for Language header tag...
            // We will use the language tag to decide what language to...
            // translate to.

            Enumeration e = req.getHeaderNames();    // cycle through HTTP headers
            while ( e.hasMoreElements() )
            {
                String header=null, value=null;      // header tag and value

                // read next header element
                try
                {
                    header=(String)e.nextElement();   // HTTP header tag name
                    value = req.getHeader(header);    // value of header tag
                    res.setHeader(header, value);     // move header into response buffer
                }
                catch ( java.lang.IllegalStateException isex )
                {
                    try
                    {
                        //Probably a date header
                        res.setDateHeader(header, req.getDateHeader(header));
                    }
                    catch ( Throwable th )
                    {
                        //Some major problem
                        header = null;
                    }
                }

                if ( header==null )
                {                  // It is likely that an exception occurred...
                    continue;                         // while reading the header, skip to next header item. 
                }

                if ( header.equalsIgnoreCase("content-type") )
                {
                    contentType=value;                 // Set content type of output buffer
                }

                //check for admin-defined language
                if ( useLang!=null )
                {
                    // The following statement checks the ltLang array to find "the best" matching
                    // language.   If the language tranlator has yet to be loaded, this method 
                    // will load it.
                    langIndex=loadLanguage(useLang);   // find language index
                }
                else if ( header.equalsIgnoreCase("ACCEPT_LANGUAGE") || 
                          header.equalsIgnoreCase("ACCEPT-LANGUAGE") )
                { // check for HTTP language tag
                    // The following statement checks the ltLang array to find "the best" matching
                    // language.   If the language tranlator has yet to be loaded, this method 
                    // will load it.
                    langIndex=loadLanguage( value );   // find language index
                }
            } // end-while

            if ( contentType==null )
            {
                contentType=req.getContentType();        // content type of request buffer
                if ( contentType==null )
                {
                    contentType="text/html";
                }
            }
            res.setContentType(contentType);      

            // Read the request buffer, translate and write to response buffer.
            int     n;                               // number of bytes to translate
            byte    bytes[]= new byte[10480];        // buffer of text to translate     
            boolean translate = false;               // translation indicator
            Object  handle = null;                   // translation context handle
            com.ibm.lt.LTinterface ltInterface=null; // lanuage translation class

            if ( (langIndex!=-1) &&                   // Translate if we have a target language...
                 (contentType.equals("text/html")||   // and our content is text (plain or html)
                  contentType.equals("text/plain")) )
            {

                // set translation indicator to "green light" 
                translate=true;                       

                // Select the language translation class corresponding to the language index
                // the language translation class knows how to translate to the given target language
                ltInterface= ltClass[langIndex];

                // Given a set of options, create a translation handle (context) for the
                // corresponding language class
                handle=ltInterface.jltBeginTranslation(format+" *path="+mtPath);
            }

            if ( translate && handle!=null )
            { // Translate if an engine loaded and contentType is correct

                byte retbytes[]=null;         // initialize translated text buffer

                while ( true )
                {                // read content from input buffer
                    n=in.read(bytes);          // Just read a stream of bytes who knows how big returned
                    if ( n==(-1) )               // use a nice buffer size of 10K max
                        break;                  // size of -1 means EOS has been reached.

                    // System.out.println("MachineTranslateFilter:  service()  ...about to translate- text="+new String(bytes,0,n));

                    retbytes=ltInterface.jltbExtTranslate(handle, new String(bytes,0,n));

                    // System.out.println("MachineTranslateFilter:  service()  ...translated- text="+new String(retbytes));

                    if ( retbytes!=null )
                    {      // Was there data to be translated?
                        out.write(retbytes);    // Then print it out
                        continue;
                    }
                }
                retbytes=ltInterface.jltbExtTranslate(handle,null);  //Signal EOF and translate the remaining
                if ( retbytes!=null )
                {         // Was there remaining data to translation?
                    out.write(retbytes);       // yep, print it out.
                }
                ltInterface.jltEndTranslation(handle);

            }
            else
            {
                // Content is not supported... copy entire file to response buffer

                while ( true )
                {               //Very Fast this way
                    n= in.read(bytes);
                    if ( n==-1 )
                        break;
                    out.write(bytes,0,n);
                }
            }
            in.close();
            out.close();

        }
        catch ( Exception e )
        {
            System.out.println("Machine Translate Filter (MTFilter): ERROR! Genaral Exception in MTFilter. "+e);
            e.printStackTrace();
        }
    }    

    /**
     * loadLanguage
     *
     * Parse the La nguage Tag String and return an index into the ltLang array.
     * The following method checks the ltLang array to find "the best" matching
     * language.   If the language tranlator has yet to be loaded, this method 
     * will load it.
     * The following is a sample ACCEPT_LANGUAGE http tag
     *          en-us,it;q=0.8,fr;q=0.6,zh-cn;q=0.4,de;q=0.2
     *
     * @param      headerTag  a HTML Header tag
     * @return     index      index into ltLang array, else -1 if nothing found
     */
    private int  loadLanguage(String qvalues)
    {
        if ( (qvalues==null) && (useLang==null) )
        {
            // System.out.println("No Language to Load!");
            return(-1);
        }
        else if ( useLang==null )
        {   //use HTTP Header Info   
            //System.out.println("Loading Language  "+qvalues);
            String string=null;
            StringTokenizer st = new StringTokenizer(qvalues,",");
            int i = -1;
            while ( (st.hasMoreTokens()) && (i==-1) )
            {
                i=loadLg(st.nextToken().trim().toLowerCase());
            }
            return i;
        }
        else
        {
            // System.out.println("Loading Language  "+useLang);
            return loadLg(useLang);
        }
    }

    private int loadLg(String lang)
    {

        Class cls;
        if ( lang.equals("zh-tw") )
            lang="tw";
        else
        {
            if ( lang.equals("zh") )
                lang="cn";
        }
        lang=lang.substring(0,2);

        String engineName="com.ibm.lt.LTen"+lang;
        try
        {
            for ( int i=0; i<ltLang.length; i++ )
            {
                if ( ltLang[i].equalsIgnoreCase(lang) )
                {
                    if ( ltClass[i]==null )
                    {
                        synchronized (ltClass)
                        {
                            if ( ltClass[i]==null )
                            {
                                cls=Class.forName(engineName);
                                ltClass[i] = (com.ibm.lt.LTinterface) cls.newInstance();
                            }
                        }
                        // System.out.println("SUCCESS: loadLanguage  "+engineName);
                    }
                    return i;
                }
            }
        }
        catch ( Throwable e )
        {
            System.out.println("Machine Translate Filter (MTFilter): ERROR: Failed to load language "+engineName);
            e.printStackTrace();
        }
        return -1;
    }

    public String getServletInfo()
    {
        return "Sample Machine Translation Filter Servlet";
    }

}
